@@ -1,7 +1,7 @@
 /* ks-engine-ldap.c - talk to a LDAP keyserver
  * Copyright (C) 2001, 2002, 2004, 2005, 2006
  *               2007  Free Software Foundation, Inc.
- * Copyright (C) 2015, 2020, 2023  g10 Code GmbH
+ * Copyright (C) 2015, 2020, 2023, 2025  g10 Code GmbH
  *
  * This file is part of GnuPG.
  *
@@ -54,6 +54,7 @@
 #define SERVERINFO_SCHEMAV2 4 /* Version 2 of the Schema.            */
 #define SERVERINFO_NTDS     8 /* Server is an Active Directory.      */
 #define SERVERINFO_GENERIC 16 /* Connected in generic mode.          */
+#define SERVERINFO_CNFPR   32 /* Server uses CN=fingerprint.         */
 
 
 /* The page size requested from the server.  */
@@ -615,6 +616,8 @@ interrogate_ldap_dn (LDAP *ldap_conn, const char *basedn_search,
   lerr = ldap_search_s (ldap_conn, object, LDAP_SCOPE_BASE,
                         "(objectClass=*)", attr2, 0, &si_res);
   npth_protect ();
+  if (DBG_LDAP)
+    log_debug ("%s: searched for '%s': ldaprc=%d\n", __func__, object, lerr);
   xfree (object);
 
   if (lerr == LDAP_SUCCESS)
@@ -623,12 +626,14 @@ interrogate_ldap_dn (LDAP *ldap_conn, const char *basedn_search,
       if (vals && vals[0])
         basedn = xtrystrdup (vals[0]);
       my_ldap_value_free (vals);
+      if (DBG_LDAP)
+        log_debug ("%s: baseDN='%s'\n", __func__, basedn);
 
       vals = ldap_get_values (ldap_conn, si_res, "pgpSoftware");
       if (vals && vals[0])
         {
-          if (opt.debug)
-            log_debug ("Server: \t%s\n", vals[0]);
+          if (DBG_LDAP)
+            log_debug ("%s: pgpSoftware: \t%s\n", __func__, vals[0]);
           if (!ascii_strcasecmp (vals[0], "GnuPG"))
             is_gnupg = 1;
         }
@@ -637,8 +642,6 @@ interrogate_ldap_dn (LDAP *ldap_conn, const char *basedn_search,
       vals = ldap_get_values (ldap_conn, si_res, "pgpVersion");
       if (vals && vals[0])
         {
-          if (opt.debug)
-            log_debug ("Version:\t%s\n", vals[0]);
           if (is_gnupg)
             {
               const char *fields[2];
@@ -648,8 +651,16 @@ interrogate_ldap_dn (LDAP *ldap_conn, const char *basedn_search,
                 *r_serverinfo |= SERVERINFO_SCHEMAV2;
               if (nfields > 1
                   && !ascii_strcasecmp (fields[1], "ntds"))
-                *r_serverinfo |= SERVERINFO_NTDS;
+                *r_serverinfo |= (SERVERINFO_NTDS|SERVERINFO_CNFPR);
+              else if (nfields > 1
+                       && !ascii_strcasecmp (fields[1], "cnfpr"))
+                *r_serverinfo |= SERVERINFO_CNFPR;
             }
+          if (DBG_LDAP)
+            log_debug ("%s: pgpVersion:\t%s%s%s%s\n", __func__, vals[0],
+                       (*r_serverinfo & SERVERINFO_SCHEMAV2)? " schema2":"",
+                       (*r_serverinfo & SERVERINFO_NTDS)? " ntds":"",
+                       (*r_serverinfo & SERVERINFO_CNFPR)? "cnfpr":"");
         }
       my_ldap_value_free (vals);
     }
@@ -754,9 +765,9 @@ my_ldap_connect (parsed_uri_t uri, unsigned int generic, LDAP **ldap_connp,
   if (opt.verbose)
     log_info ("ldap connect to '%s:%d:%s:%s:%s:%s%s%s'%s\n",
               host, port,
-              basedn_arg ? basedn_arg : "",
               bindname ? bindname : "",
               password ? "*****" : "",
+              basedn_arg ? basedn_arg : "",
               use_tls == 1? "starttls" : use_tls == 2? "ldaptls" : "plain",
               use_ntds ? ",ntds":"",
               use_areconly? ",areconly":"",
@@ -906,11 +917,13 @@ my_ldap_connect (parsed_uri_t uri, unsigned int generic, LDAP **ldap_connp,
       npth_protect ();
       if (lerr != LDAP_SUCCESS)
 	{
-	  log_error ("error binding to LDAP via AD: %s\n",
+	  log_error ("error binding to LDAP via NTDS: %s\n",
                      ldap_err2string (lerr));
           err = ldap_err_to_gpg_err (lerr);
 	  goto out;
 	}
+      if (DBG_LDAP)
+        log_debug ("%s: ldap_bind to NTDS succeeded\n", __func__);
 #else
       log_error ("ldap: no Active Directory support but 'ntds' requested\n");
       err = gpg_error (GPG_ERR_NOT_SUPPORTED);
@@ -928,10 +941,14 @@ my_ldap_connect (parsed_uri_t uri, unsigned int generic, LDAP **ldap_connp,
           err = ldap_err_to_gpg_err (lerr);
 	  goto out;
 	}
+      if (DBG_LDAP)
+        log_debug ("%s: ldap_bind to '%s' succeeded\n", __func__, bindname);
     }
   else
     {
       /* By default we don't bind as there is usually no need to.  */
+      if (DBG_LDAP)
+        log_debug ("%s: ldap_bind not used\n", __func__);
     }
 
   if (generic)
@@ -947,6 +964,9 @@ my_ldap_connect (parsed_uri_t uri, unsigned int generic, LDAP **ldap_connp,
               goto out;
             }
         }
+      if (DBG_LDAP)
+        log_debug ("%s: serverinfo set to generic; basedn '%s'\n", __func__,
+                   basedn);
     }
   else if (basedn_arg && *basedn_arg)
     {
@@ -967,6 +987,9 @@ my_ldap_connect (parsed_uri_t uri, unsigned int generic, LDAP **ldap_connp,
             basedn = interrogate_ldap_dn (ldap_conn, basedn_parent + 1,
                                           r_serverinfo);
         }
+      if (DBG_LDAP)
+        log_debug ("%s: serverinfo set to realldap; basedn '%s'\n", __func__,
+                   basedn);
     }
   else
     { /* Look for namingContexts.  */
@@ -977,6 +1000,8 @@ my_ldap_connect (parsed_uri_t uri, unsigned int generic, LDAP **ldap_connp,
       lerr = ldap_search_s (ldap_conn, "", LDAP_SCOPE_BASE,
 			   "(objectClass=*)", attr, 0, &res);
       npth_protect ();
+      if (DBG_LDAP)
+        log_debug ("%s: searched namingContexts: lerr=%d\n", __func__, lerr);
 
       if (lerr == LDAP_SUCCESS)
 	{
@@ -1000,7 +1025,15 @@ my_ldap_connect (parsed_uri_t uri, unsigned int generic, LDAP **ldap_connp,
                                               r_serverinfo);
 
 	      ldap_value_free (context);
+              if (DBG_LDAP)
+                log_debug ("%s: namingContext found (realldap); basedn='%s'\n",
+                           __func__, basedn);
 	    }
+          else
+            {
+              if (DBG_LDAP)
+                log_debug ("%s: namingContext not found\n", __func__);
+            }
 	}
       else /* ldap_search failed.  */
 	{
@@ -1015,6 +1048,9 @@ my_ldap_connect (parsed_uri_t uri, unsigned int generic, LDAP **ldap_connp,
 	  lerr = ldap_search_s (ldap_conn, "cn=pgpServerInfo", LDAP_SCOPE_BASE,
 			       "(objectClass=*)", attr2, 0, &si_res);
           npth_protect ();
+          if (DBG_LDAP)
+            log_debug ("%s: searched cn=pgpServerInfo: lerr=%d\n",
+                       __func__, lerr);
 	  if (lerr == LDAP_SUCCESS)
 	    {
 	      /* For the PGP LDAP keyserver, this is always
@@ -1025,6 +1061,8 @@ my_ldap_connect (parsed_uri_t uri, unsigned int generic, LDAP **ldap_connp,
 	      if (vals && vals[0])
 		{
 		  basedn = xtrystrdup (vals[0]);
+                  if (DBG_LDAP)
+                    log_debug ("%s: baseKeySpaceDN='%s'\n", __func__, basedn);
 		}
               my_ldap_value_free (vals);
 
@@ -1032,7 +1070,8 @@ my_ldap_connect (parsed_uri_t uri, unsigned int generic, LDAP **ldap_connp,
 	      if (vals && vals[0])
 		{
                   if (opt.debug)
-                    log_debug ("ks-ldap: PGP Server: \t%s\n", vals[0]);
+                    log_debug ("%s: PGP Server software='%s'\n",
+                               __func__, vals[0]);
 		}
               my_ldap_value_free (vals);
 
@@ -1040,7 +1079,8 @@ my_ldap_connect (parsed_uri_t uri, unsigned int generic, LDAP **ldap_connp,
 	      if (vals && vals[0])
 		{
                   if (opt.debug)
-                    log_debug ("ks-ldap: PGP Server Version:\t%s\n", vals[0]);
+                    log_debug ("%s: PGP Server version='%s'\n",
+                               __func__, vals[0]);
 
 		  /* If the version is high enough, use the new
 		     pgpKeyV2 attribute.  This design is iffy at best,
@@ -1065,7 +1105,7 @@ my_ldap_connect (parsed_uri_t uri, unsigned int generic, LDAP **ldap_connp,
     }
 
  out:
-  if (!err && opt.debug)
+  if (!err && DBG_LDAP)
     {
       log_debug ("ldap_conn: %p\n", ldap_conn);
       log_debug ("server_type: %s\n",
@@ -1814,7 +1854,7 @@ ks_ldap_get (ctrl_t ctrl, parsed_uri_t uri, const char *keyspec,
           filter = fstr;
         }
 
-      if (opt.debug)
+      if (DBG_LDAP)
         log_debug ("ks-ldap: using filter: %s\n", filter);
 
       /* Replace "dummy".  */
@@ -1974,7 +2014,7 @@ ks_ldap_search (ctrl_t ctrl, parsed_uri_t uri, const char *pattern,
         NULL
       };
 
-    if (opt.debug)
+    if (DBG_LDAP)
       log_debug ("SEARCH '%s' => '%s' BEGIN\n", pattern, filter);
 
     npth_unprotect ();
@@ -2183,7 +2223,7 @@ ks_ldap_search (ctrl_t ctrl, parsed_uri_t uri, const char *pattern,
     free_strlist (dupelist);
   }
 
-  if (opt.debug)
+  if (DBG_LDAP)
     log_debug ("SEARCH %s END\n", pattern);
 
  out:
@@ -2978,7 +3018,7 @@ ks_ldap_put (ctrl_t ctrl, parsed_uri_t uri,
     char **attrval;
     char *dn;
 
-    if ((serverinfo & SERVERINFO_NTDS))
+    if ((serverinfo & SERVERINFO_CNFPR))
       {
         /* The modern way using a CN RDN with the fingerprint.  This
          * has the advantage that we won't have duplicate 64 bit
@@ -3011,7 +3051,7 @@ ks_ldap_put (ctrl_t ctrl, parsed_uri_t uri,
         err = gpg_error_from_syserror ();
         goto out;
       }
-    if (opt.debug)
+    if (DBG_LDAP)
       log_debug ("ks-ldap: using DN: %s\n", dn);
 
     npth_unprotect ();
@@ -3048,15 +3088,113 @@ ks_ldap_put (ctrl_t ctrl, parsed_uri_t uri,
 }
 
 
-/* Delete the keys given by PATTERNS from the keyserver identified by
- * URI. */
+/* Delete the keys given by the list of fingerprints in FPRLIST from
+ * the keyserver identified by URI.  The function stops at the first
+ * error encountered. */
 gpg_error_t
-ks_ldap_del (ctrl_t ctrl, parsed_uri_t uri, strlist_t patterns)
+ks_ldap_del (ctrl_t ctrl, parsed_uri_t uri, strlist_t fprlist)
 {
-  (void)ctrl;
-  (void)uri;
-  (void)patterns;
-  return gpg_error (GPG_ERR_NOT_IMPLEMENTED);
+  gpg_error_t err = 0;
+  int ldap_err;
+  unsigned int serverinfo;
+  LDAP *ldap_conn = NULL;
+  char *basedn = NULL;
+  char *dn = NULL;
+  strlist_t fpr;
+  unsigned int count = 0;
+  unsigned int totalcount = 0;
+
+  if (dirmngr_use_tor ())
+    {
+      return no_ldap_due_to_tor (ctrl);
+    }
+
+  for (fpr = fprlist; fpr; fpr = fpr->next)
+    totalcount++;
+
+  err = my_ldap_connect (uri, 0, &ldap_conn, &basedn, NULL, NULL, &serverinfo);
+  if (err || !basedn)
+    {
+      if(opt.verbose)
+        log_info ("%s: connecting to server failed\n", __func__);
+      if (!err)
+	err = gpg_error (GPG_ERR_GENERAL);  /* (no baseDN) */
+      goto leave;
+    }
+
+  if (!(serverinfo & SERVERINFO_REALLDAP))
+    {
+      if(opt.verbose)
+        log_info ("%s: The PGP.com keyserver is not supported\n", __func__);
+      err = gpg_error (GPG_ERR_NOT_SUPPORTED);
+      goto leave;
+    }
+
+  if (!(serverinfo & SERVERINFO_SCHEMAV2))
+    {
+      if(opt.verbose)
+        log_info ("%s: The keyserver does not support the v2 schema\n",
+                  __func__);
+      err = gpg_error (GPG_ERR_NOT_SUPPORTED);
+      goto leave;
+    }
+
+  if (opt.verbose)
+    log_info ("%s: Using DN: %s,%s\n", __func__,
+              (serverinfo & SERVERINFO_CNFPR)? "CN=<fingerprint>"
+              /* */                          : "pgpCertID=<keyid>",
+              basedn);
+  for (fpr = fprlist; fpr; fpr = fpr->next)
+    {
+      if ((serverinfo & SERVERINFO_CNFPR))
+        {
+          xfree (dn);
+          dn = xtryasprintf ("CN=%s,%s", fpr->d, basedn);
+        }
+      else
+        {
+          unsigned int off;
+
+          /* Simle method to get the keyID.  Note that a v5 key
+           * (len>40) has the keyid at the left.  If the length is
+           * less than 17 we assume a keyid has been given.  */
+          off = strlen (fpr->d);
+          if (off <= 40 && off > 16)
+            off = off - 16;
+          else
+            off = 0;
+
+          xfree (dn);
+          dn = xtryasprintf ("pgpCertID=%.16s,%s", fpr->d+off, basedn);
+        }
+
+      npth_unprotect ();
+      ldap_err = ldap_delete_ext_s (ldap_conn, dn, NULL, NULL);
+      npth_protect ();
+      if (ldap_err == LDAP_SUCCESS)
+        {
+          if (opt.verbose)
+            log_info ("%s: key %s deleted\n", __func__, fpr->d);
+          count++;
+        }
+      else
+        {
+          log_error ("%s: error deleting key %s: %s\n",
+                     __func__, fpr->d, ldap_err2string (ldap_err));
+          err = ldap_err_to_gpg_err (ldap_err);
+          break;  /* Stop at the first failed deletion.  */
+        }
+    }
+  log_info ("%s: number of keys deleted: %u of %u\n",
+            __func__, count, totalcount);
+
+
+ leave:
+  if (ldap_conn)
+    ldap_unbind (ldap_conn);
+  xfree (dn);
+  xfree (basedn);
+  return err;
 }
 
 
@@ -3226,7 +3364,7 @@ ks_ldap_query (ctrl_t ctrl, parsed_uri_t uri, unsigned int ks_get_flags,
           basedn = basedn_from_rootdse (ctrl, uri);
         }
 
-      if (opt.debug)
+      if (DBG_LDAP)
         {
           log_debug ("ks-ldap: using basedn: %s\n", basedn);
           log_debug ("ks-ldap: using filter: %s\n", filter);
