@@ -1,5 +1,5 @@
 /* call-keyboxd.c - Access to the keyboxd storage server
- * Copyright (C) 2019  g10 Code GmbH
+ * Copyright (C) 2019, 2024  g10 Code GmbH
  *
  * This file is part of GnuPG.
  *
@@ -38,7 +38,6 @@
 #include "../common/i18n.h"
 #include "../common/asshelp.h"
 #include "../common/host2net.h"
-#include "../common/exechelp.h"
 #include "../common/status.h"
 #include "../kbx/kbx-client-util.h"
 #include "keydb.h"
@@ -47,7 +46,7 @@
 
 
 /* Data used to keep track of keybox daemon sessions.  This allows us
- * to use several sessions with the keyboxd and also to re-use already
+ * to use several sessions with the keyboxd and also to reuse already
  * established sessions.  Note that gpg.h defines the type
  * keyboxd_local_t for this structure. */
 struct keyboxd_local_s
@@ -94,8 +93,6 @@ gpg_keyboxd_deinit_session_data (ctrl_t ctrl)
         log_error ("oops: trying to cleanup an active keyboxd context\n");
       else
         {
-          kbx_client_data_release (kbl->kcd);
-          kbl->kcd = NULL;
           if (kbl->ctx && in_transaction)
             {
               /* This is our hack to commit the changes done during a
@@ -106,12 +103,21 @@ gpg_keyboxd_deinit_session_data (ctrl_t ctrl)
               err = assuan_transact (kbl->ctx, "TRANSACTION commit",
                                      NULL, NULL, NULL, NULL, NULL, NULL);
               if (err)
-                log_error ("error commiting last transaction: %s\n",
+                log_error ("error committing last transaction: %s\n",
                             gpg_strerror (err));
               in_transaction = 0;
             }
           assuan_release (kbl->ctx);
           kbl->ctx = NULL;
+          /*
+           * Since there may be pipe output FD sent to the server (so
+           * that it can receive data through the pipe), we should
+           * release the assuan connection before releasing KBL->KCD.
+           * This way, the data receiving thread can finish cleanly,
+           * and we can join the thread.
+           */
+          kbx_client_data_release (kbl->kcd);
+          kbl->kcd = NULL;
         }
       xfree (kbl);
     }
@@ -143,7 +149,8 @@ create_new_context (ctrl_t ctrl, assuan_context_t *r_ctx)
   err = start_new_keyboxd (&ctx,
                            GPG_ERR_SOURCE_DEFAULT,
                            opt.keyboxd_program,
-                           opt.autostart, opt.verbose, DBG_IPC,
+                           opt.autostart?ASSHELP_FLAG_AUTOSTART:0,
+                           opt.verbose, DBG_IPC,
                            NULL, ctrl);
   if (!opt.autostart && gpg_err_code (err) == GPG_ERR_NO_KEYBOXD)
     {
@@ -223,7 +230,7 @@ open_context (ctrl_t ctrl, keyboxd_local_t *r_kbl)
           return err;
         }
 
-      err = kbx_client_data_new (&kbl->kcd, kbl->ctx, 1);
+      err = kbx_client_data_new (&kbl->kcd, kbl->ctx, 0);
       if (err)
         {
           assuan_release (kbl->ctx);
@@ -385,6 +392,23 @@ keydb_get_keyblock (KEYDB_HANDLE hd, kbnode_t *ret_kb)
 }
 
 
+/* Default status callback used to show diagnostics from the keyboxd  */
+static gpg_error_t
+keydb_default_status_cb (void *opaque, const char *line)
+{
+  const char *s;
+
+  (void)opaque;
+
+  if ((s = has_leading_keyword (line, "NOTE")))
+    log_info (_("Note: %s\n"), s);
+  else if ((s = has_leading_keyword (line, "WARNING")))
+    log_info (_("WARNING: %s\n"), s);
+
+  return 0;
+}
+
+
 
 /* Communication object for STORE commands.  */
 struct store_parm_s
@@ -464,7 +488,8 @@ keydb_update_keyblock (ctrl_t ctrl, KEYDB_HANDLE hd, kbnode_t kb)
   err = assuan_transact (hd->kbl->ctx, "STORE --update",
                          NULL, NULL,
                          store_inq_cb, &parm,
-                         NULL, NULL);
+                         keydb_default_status_cb, hd);
+
 
  leave:
   iobuf_close (iobuf);
@@ -515,7 +540,7 @@ keydb_insert_keyblock (KEYDB_HANDLE hd, kbnode_t kb)
   err = assuan_transact (hd->kbl->ctx, "STORE --insert",
                          NULL, NULL,
                          store_inq_cb, &parm,
-                         NULL, NULL);
+                         keydb_default_status_cb, hd);
 
  leave:
   iobuf_close (iobuf);
@@ -561,7 +586,7 @@ keydb_delete_keyblock (KEYDB_HANDLE hd)
   err = assuan_transact (hd->kbl->ctx, line,
                          NULL, NULL,
                          NULL, NULL,
-                         NULL, NULL);
+                         keydb_default_status_cb, hd);
 
  leave:
   return err;
@@ -605,7 +630,7 @@ keydb_search_reset (KEYDB_HANDLE hd)
 
 
 
-/* Status callback for SEARCH and NEXT operaions.  */
+/* Status callback for SEARCH and NEXT operations.  */
 static gpg_error_t
 search_status_cb (void *opaque, const char *line)
 {
@@ -648,6 +673,8 @@ search_status_cb (void *opaque, const char *line)
             }
         }
     }
+  else
+    err = keydb_default_status_cb (opaque, line);
 
   return err;
 }
@@ -690,7 +717,7 @@ keydb_search (KEYDB_HANDLE hd, KEYDB_SEARCH_DESC *desc,
   if (DBG_CLOCK)
     log_clock ("%s enter", __func__);
 
-  if (DBG_LOOKUP)
+  if (DBG_KEYDB)
     {
       log_debug ("%s: %zu search descriptions:\n", __func__, ndesc);
       for (i = 0; i < ndesc; i ++)
@@ -759,7 +786,8 @@ keydb_search (KEYDB_HANDLE hd, KEYDB_SEARCH_DESC *desc,
           break;
 
         case KEYDB_SEARCH_MODE_MAIL:
-          snprintf (line, sizeof line, "SEARCH %s -- <%s", more, desc->u.name);
+          snprintf (line, sizeof line, "SEARCH %s -- <%s",
+                    more, desc->u.name+(desc->u.name[0] == '<') );
           break;
 
         case KEYDB_SEARCH_MODE_MAILSUB:
@@ -852,7 +880,7 @@ keydb_search (KEYDB_HANDLE hd, KEYDB_SEARCH_DESC *desc,
     {
       hd->kbl->search_result = iobuf_temp_with_content (buffer, len);
       xfree (buffer);
-      if (DBG_LOOKUP && hd->last_ubid_valid)
+      if (DBG_KEYDB && hd->last_ubid_valid)
         log_printhex (hd->last_ubid, 20, "found UBID (%d,%d):",
                       hd->last_uid_no, hd->last_pk_no);
     }
